; ------------------------------------------------------------------------ ;
; LZW Decoder
; Author: Jack Thornton
; CetaSoft / Level Five, Inc.
; ------------------------------------------------------------------------ ;

	IDEAL
	P486
	MODEL USE32 SMALL

; ************************************************************** DECLARATIONS
; C/C++ binding macros

    ; helper macros for making void 'C' routines
    ; includes prolog and epilog code
    MACRO CROUTINE name
	PROC _&name near
	PUBLIC _&name
	pushad
	mov	ebp,esp
    ENDM

    MACRO ENDCROUTINE name
	popad
	ret
	ENDP _&name
    ENDM

; paramater base - offset of first passed parameter (from EBP)
PARAMBASE   EQU  36	; 32 bytes PUSHAD + 4 bytes return address

; data types

	STRUC DictNode		; 5-bytes per entry
	    Length	DW ?
	    Parent	DW ?
	    Byte	DB ?
	ENDS DictNode

; constant declarations

	; special-case compression codes
	END_OF_STREAM	EQU 256
	BUMP_CODE	EQU 257
	FLUSH_CODE	EQU 258
	FIRST_CODE	EQU 259

	; error codes
	ERR_ASSERTION	EQU 1
	ERR_BADPARAMS	EQU 2
	ERR_SRCOVERRUN	EQU 5
	ERR_DESTOVERRUN EQU 6

; ********************************************************** INITIALIZED DATA
	DATASEG

; ******************************************************** UNINITIALIZED DATA
	UDATASEG

	; error code (for reporting to DLL client)
	_curError	DD ?	; current error code

	; misc. decompression variables
	_newCode	DD ?	; newest code extracted from source stream
	_character	DD ?	; newest/last explicit character from source

	; source and destination buffer lengths (for error checking)
	_srcLength	DD ?	; length of source data
	_destLength	DD ?	; length of destination buffer

	; variables used for parsing source bitstream
	_bitPos 	DD ?	; current bit position within source stream
	_excessBitLength DD ?	; 32 - codeBitLength (for rotating)

	; dictionary maint. variables
	_nextCode	DD ?	; next entry to assign in dictionary
	_codeBitLength	DD ?	; current length of codes (in bits)
	_nextBumpCode	DD ?	; next code at which we must increase length

	; the dictionary
	_dictionary	DictNode 4096 dup(?)

	; allow C++ stuff to access these variables
	public _curError

	; make these public for debugging purposes (visible to debugger)
	public _dictionary
	public _destLength,_srcLength,_nextBumpCode,_bitPos
	public _excessBitLength,_codeBitLength,_character,_newCode,_nextCode

	; table entry member equates
	DICTENTRY_LENGTH     EQU offset _dictionary
	DICTENTRY_PARENT     EQU offset _dictionary + 2
	DICTENTRY_CHAR	     EQU offset _dictionary + 4

; ********************************************************************** CODE
	CODESEG

; GET_CODE
;   This macro grabs a collection of bits from the incoming bitstream
;   and presents them right-aligned in EAX. Codes in this LZW scheme
;   are variable-length, from a minimum of 9 bits to a maximum of 12 bits.
;
    MACRO GET_CODE
	; get current bit position within source stream
	mov	edx,[_bitPos]

	; read a collection of bits from the source stream
	mov	eax,[esi]

	; rotate bits within EAX to account for code length and starting pos
	bswap	eax
	mov	ecx,[_excessBitLength]
	sub	ecx,edx
	ror	eax,cl

	; mask off the higher, unwanted bits
	and	eax,[_nextBumpCode]	; bump code doubles as bitmask

	; advance source bitstream pointer (bit position and byte position)
	add	edx,[_codeBitLength]	; add bit length to current position
	mov	ecx,edx 		; divide new _bitPos by 8 to get number
	shr	ecx,3			; of bytes to advance src pointer (esi)
	and	edx,7			; saved _bitPos is then newpos mod 8
	add	esi,ecx 		; add byte advance
	mov	[_bitPos],edx		; and save position of next code
    ENDM

;
; ExpandFile
;
;   'C' prototype:
;	void ExpandFile(void *dest, void *src, int destLen, int srcLen);
;
;   The file expander operates much like the encoder.  It has to
; read in codes, the convert the codes to a string of characters.
; The only catch in the whole operation occurs when the encoder
; encounters a CHAR+STRING+CHAR+STRING+CHAR sequence.  When this
; occurs, the encoder outputs a code that is not presently defined
; in the table.  This is handled as an exception.  All of the special
; input codes are handled in various ways.
;
; The following registers hold consistant values throughout the code:
;	EBP:	old/parent code (after parameters are pulled off the stack)
;	ESI:	source data pointer (input codes)
;	EDI:	destination buffer pointer (output data)
;
; All registers are preserved as a result of the CROUTINE/ENDCROUTINE
; macros. This routine returns nothing to the caller (void).
;
; --------------------------------------------------------------------------
    CROUTINE ExpandFile

	; pull the parameters off the stack
	mov	edi,[ebp+PARAMBASE]	    ; destination pointer into EDI
	mov	esi,[ebp+PARAMBASE+4]	    ; source pointer into ESI
	mov	eax,[ebp+PARAMBASE+8]
	mov	[_destLength],eax
	mov	eax,[ebp+PARAMBASE+12]
	mov	[_srcLength],eax

	; initialize bit position within source stream
	mov	[_bitPos],0

	; initialize direction flag autoincrements
	cld

	; initialize the first 256 entries in the dictionary (fixed portion)
	mov	ecx,255
atoms:	mov	[word ptr ecx+ecx*4+DICTENTRY_LENGTH],1
	mov	[word ptr ecx+ecx*4+DICTENTRY_PARENT],0
	mov	[byte ptr ecx+ecx*4+DICTENTRY_CHAR],cl
	dec	ecx
	jns	atoms

FlushDictionary:
	; initialize the non-atomic portions of dictionary (clear to zero)
	mov	ebx,edi 		    ; save destination buffer ptr
	mov	ecx,5*(4096-256)/4	    ; size of dictionary, in DWORDS
	mov	edi,offset _dictionary+(5*256)	; point to dictionary
	xor	eax,eax 		    ; zero EAX
	rep	stosd			    ; blast it clean!
	mov	edi,ebx 		    ; restore destination ptr

	; initialize dictionary variables
	mov	[_nextCode],FIRST_CODE
	mov	[_codeBitLength],9
	mov	[_excessBitLength],23	    ; 32-bit length minus 9-bit codelen
	mov	[_nextBumpCode],511

	; get the initial code from the stream
	GET_CODE

	; if it is "end-of-stream", we are done
	cmp	eax,END_OF_STREAM
	je	success

	; if not, stash it in "character" and "oldCode" (ebp = oldCode)
	mov	ebp,eax
	mov	[_character],eax

	; output the code
	stosb

	; error testing - check length
	dec	[_destLength]
	jz	ErrDestOverrun

DecompressLoop:
	; get next code from the input stream
	GET_CODE

	; if end-of-stream, bail!
	cmp	eax,END_OF_STREAM
	je	success

	; if flush code, exit inner loop to reinit dictionary
	cmp	eax,FLUSH_CODE
	je	FlushDictionary

	; if it is the "bump" code, then just increase bit length
	cmp	eax,BUMP_CODE
	jne	short lookup
	inc	[_codeBitLength]    ; increment code bit length
	dec	[_excessBitLength]
	shl	[_nextBumpCode],1   ; and add another bit to the mask
	or	[_nextBumpCode],1
	jmp	DecompressLoop

	; stash the newest code
lookup: mov	[_newCode],eax

	; is this a special case or not?
	mov	edx,eax
	cmp	eax,[_nextCode]     ; if (_newCode >= _nextCode)...
	jb	short getlen
	mov	edx,ebp 	    ; ...then decode from old code instead
	or	eax,80000000h	    ; and set flag (sign bit) for later

	; look up length of the resultant string in the dictionary
getlen: movzx	ecx,[word ptr edx+edx*4+DICTENTRY_LENGTH]

	; for the sake of robustness... check for destination overrun
	sub	[_destLength],ecx
	js	short ErrDestOverrun

	; move destination pointer to end of string & set up for decrement
	lea	edi,[edi+ecx-1]
	mov	ebx,edi 	    ; save this address for later restoral
	std			    ; auto decrement instead of increment

StringLoop:
	; lookup and output character from code table entry
	mov	al,[byte ptr edx+edx*4+DICTENTRY_CHAR]
	stosb

	; get parent code into edx
	mov	edx,[dword ptr edx+edx*4+DICTENTRY_PARENT]
	and	edx,0fffh

	; decrement loop counter
	loop	StringLoop

	; save "first" character of string to store into the new code
	mov	ah,al

	; restore destination pointer past end of string we just copied
	lea	edi,[ebx+1]
	cld			    ; and restore direction flag to inc.

	; finish special-case handling (shovel out copy of last character)
	or	eax,eax 	    ; check special-case flag
	jns	short AddEntry
	mov	al,[byte ptr _character]    ; yes: one more character to write
	stosb

AddEntry:
	; make new entry in the table for this string
	mov	edx,[_nextCode] 	 ; get table index
	mov	[byte ptr edx+edx*4+DICTENTRY_CHAR],ah	    ; character
	mov	[word ptr edx+edx*4+DICTENTRY_PARENT],bp    ; old code
	mov	ebx,[dword ptr ebp+ebp*4+DICTENTRY_LENGTH]
	inc	ebx
	mov	[word ptr edx+edx*4+DICTENTRY_LENGTH],bx    ; parent length + 1

	; increment table entry for next "new" code
	inc	[_nextCode]

	; and move "_newCode" into "oldCode"
	mov	ebp,[_newCode]

	; stash character
	mov	[byte ptr _character],ah

	; continue looping until we fall out properly
	jmp	DecompressLoop

;ErrAssertion:
;	 mov	 eax,ERR_ASSERTION
;	 jmp	 short failure

ErrBadParams:
	mov	eax,ERR_BADPARAMS
	jmp	short failure

ErrSrcOverrun:
	mov	eax,ERR_SRCOVERRUN
	jmp	short failure

ErrDestOverrun:
	mov	eax,ERR_DESTOVERRUN

failure:
	; set the error code
	mov	[_curError],eax

	; and return FALSE
	xor	eax,eax 	; return FALSE (0) if not successful
	jmp	short done

success:
	mov	eax,1		; return TRUE if successful
done:
    ENDCROUTINE ExpandFile

	END

