#include "fwl5.h"
#include <assert.h>
#include <stdio.h>
#include "cstring.h"
#include "cpnodes.h"
#include "cpnet.h"
#include "options.h"
#include "frsrc.h"
#include "fstream.h"

//----------------------------------------------------------------------------
CPNetwork::CPNetwork()
  : firstNode(0)
  , fileData(0)
  , fileSize(0)
{
}
//----------------------------------------------------------------------------
CPNetwork::~CPNetwork()
{
  if (fileData) delete [] fileData;
}
//----------------------------------------------------------------------------
// GetFirstNode:
// The first node in the network technically is a CPC_TERMINAL with one link
// and no backlinks. This function returns the first clip you find starting
// from the terminal node.
CPNode* CPNetwork::GetFirstNode() // return first clip node.
{
  CPNode* aNode = firstNode;
  while (aNode->nodeType != CPN_FILM)
  {
    assert(aNode->data.control.controlType == CPC_TERMINAL);
    assert(aNode->links == 1);
    assert(aNode->backLinks == 0);
    aNode = aNode->pLinks[0];
  }
  return aNode;
}
//----------------------------------------------------------------------------
// ReadFromFile:
// Read the network from the named file, convert the file offsets to addresses
// and in this provisional version, read the clipnames and soundnames in so
// we can access the files.
int                                   // non-zero if successful
  CPNetwork::ReadFromFile
    (const string& fn)                // filename
{
  FILE* fp = fopen(fn.c_str(), "rb");
  if (fp != 0)
  {
    fread(&fileSize, sizeof(fileSize), 1, fp);
    if (fileData) delete [] fileData;
    fileData = new BYTE[fileSize+4];
    *(long*)fileData = fileSize;
    fread(fileData+4, 1, fileSize-4, fp);
    if (ReadFromRAM(fileData, fileSize))
    {
      File2Vector(fp, clipNames);
      File2Vector(fp, voxNames);
      return 1;
    }
  }
  return 0;
}
//----------------------------------------------------------------------------
// A few functions to make writing binary a little easier.
static void Write(ofstream& s, int v)
{
  s.write((char*)&v, sizeof(v));
}
static void Write(ofstream& s, unsigned v)
{
  s.write((char*)&v, sizeof(v));
}
static void Read(ifstream& s, int& p)
{
  s.read((char*)&p, sizeof(int));
}
//----------------------------------------------------------------------------
// WalkAllNodes:
// Walk down all the nodes.
// Stopping criterion established by remembering the address of each visited
// node.
// Const and non-const versions. We like const.
static BYTE* where;
void CPNetwork::WalkAllNodes(int (CPNetwork::*fn)(CPNode&,void*), void* user)
{
  where = new BYTE[fileSize];
  memset(where, 0, fileSize);
  Walk(firstNode, fn, user);
  delete [] where;
}
void CPNetwork::WalkAllNodes(int (CPNetwork::*fn)(CPNode&,void*), void* user) const
{
  where = new BYTE[fileSize];
  memset(where, 0, fileSize);
  Walk(firstNode, fn, user);
  delete [] where;
}
//----------------------------------------------------------------------------
// Walk:
// If the current node is not visited, call the function, and 
// recurse to forward and backward children.
// Forward children: outgoing personality.
// Backward children: not well schooled.
// Const and non-const versions to support const and non-const caller.
// The called function doesn't have to be const!
void CPNetwork::Walk(CPNode* n,int (CPNetwork::*fn)(CPNode&,void*), void* user)
{
  if (!where[((BYTE*)n)-fileData])
  {
    if ((this->*fn)(*n, user))
    {
      where[((BYTE*)n)-fileData] =1;
      int i;
      for (i=0; i<n->links; i++)
      {
        Walk(n->pLinks[i], fn, user);
      }
      for (i=0; i<n->backLinks; i++)
      {
        Walk(n->pBackLinks[i], fn, user);
      }
    }
  }
}
//----------------------------------------------------------------------------
void CPNetwork::Walk(CPNode* n,int (CPNetwork::*fn)(CPNode&,void*), void* user) const
{
  if (!where[((BYTE*)n)-fileData])
  {
    if ((this->*fn)(*n, user))
    {
      where[((BYTE*)n)-fileData] =1;
      int i;
      for (i=0; i<n->links; i++)
      {
        Walk(n->pLinks[i], fn, user);
      }
      for (i=0; i<n->backLinks; i++)
      {
        Walk(n->pBackLinks[i], fn, user);
      }
    }
  }
}
//----------------------------------------------------------------------------
int CPNetwork::SaveACount(CPNode& n, void* user)
{
  ofstream* out = (ofstream*)user;
  Write(*out, ((BYTE*)&n)-fileData);
  switch (n.nodeType)
  {
  case CPN_FILM:
    Write(*out, n.data.film.visits);
    break;
  case CPN_CONTROL:
    Write(*out, n.data.control.visitBits);
    break;
  }
  return 1;
}
//----------------------------------------------------------------------------
void CPNetwork::SaveCounts(ofstream& out) const
{
  WalkAllNodes(SaveACount, &out);
  Write(out, (int)0);
}
//----------------------------------------------------------------------------
int CPNetwork::ZeroACount(CPNode& n, void*)
{
  switch (n.nodeType)
  {
  case CPN_FILM:
    n.data.film.visits = 0;
    n.data.film.viewed = FALSE;
    break;
  case CPN_CONTROL:
    n.data.control.visitBits =0;
    break;
  }
  return 1;
}
//----------------------------------------------------------------------------
void CPNetwork::ZeroCounts()
{
  WalkAllNodes(ZeroACount, 0);
}
//----------------------------------------------------------------------------
void CPNetwork::ReadCounts(ifstream& in)
{
  CPNode* p;
  int pa;
  Read(in, pa);
  while (pa != 0)
  {
    p = (CPNode*)(pa + fileData);
    Read(in, pa);
    switch (p->nodeType)
    {
    case CPN_FILM:
      p->data.film.visits = pa;
      p->data.film.viewed = !!pa;
      break;
    case CPN_CONTROL:
      p->data.control.visitBits = pa;
      break;
    }
    Read(in, pa);
  }  
}
//----------------------------------------------------------------------------
// ReadFromFile:
// Read the network from the named file, convert the file offsets to addresses
// and in this provisional version, read the clipnames and soundnames in so
// we can access the files.
int                                   // non-zero if successful
  CPNetwork::ReadFromArchive
    (FRSRC* bigfile
    , int id
    )
{
  long handle = bigfile->OpenFileEntry(id);
  if (handle != -1)
  {
    bigfile->Read(handle, &fileSize, sizeof(fileSize));
    if (fileData) delete [] fileData;
    fileData = new BYTE[fileSize+4];
    *(long*)fileData = fileSize;
    bigfile->Read(handle, fileData+4, fileSize-4);
    if (ReadFromRAM(fileData, fileSize))
    {
      Arc2Vector(bigfile, handle, clipNames);
      Arc2Vector(bigfile, handle, voxNames);
      bigfile->CloseFile(handle);
      return 1;
    }
    bigfile->CloseFile(handle);
  }
  return 0;
}
//----------------------------------------------------------------------------
// ReadFromRam:
// Unflatten all the pointers in the network data.
int                                   // nz if successful.
  CPNetwork::ReadFromRAM
    ( BYTE* mem                       // pointer to network data
    , int bytes                       // size of network data.
    )
{
  fileData = mem;
  fileSize = bytes;
  long oFirstNode = *(long*)(fileData+4);
  *(long*)(fileData+4) = (long)(fileData+oFirstNode);
  CPNode** pFirstNode = (CPNode**)(fileData+4);
  firstNode = *pFirstNode;
  UnFlatten(firstNode);
  return 1;
}

#ifdef _DEBUG
#define B 0xffffffff
static int breakpointClipIds[8] = 
{ B,B,B,B,B,B,B,B };
#undef B
#endif
//----------------------------------------------------------------------------
// GetCompatibleLink:
// Trace a link. Stop when it points to a film.
// If it doesn't point to a film, use the psych value and any other
// criteria stipulated by the controls encountered to choose links to
// follow. 
// Each control will have one and only one link that is best.
// Factors that must be considered include:
//   How many times a specific node has been visited
//   Whether or not another node has been visited
//   Which links have been traversed from a given node
// The caller must, in some cases, predict the psych value that must
// be given to this func.
// If final is non-zero, counts will count and burns will burn.
// Otherwise it is assumed that this is called for prediction in order
// to arrange preloading of leadins, so the data are not changed.
//
CPNode*                               // The link to be used
  CPNetwork::GetCompatibleLink
    ( CPNode* iLink                   // The link encountered
    , const Psych& psych              // the psych value to match with
    , int final                       // TRUE (defaults to false) if links burn.
    ) const
{
  extern GameOptions globalOptions;
  CPNode* rLink = iLink;
  CPNode* tLink;
  unsigned k;
  int i;
  unsigned d;
#ifdef _DEBUG
  if (iLink->nodeType == CPN_FILM)
  {
    for (i=0; i<8; i++)
    {
      if (iLink->data.film.clip == breakpointClipIds[i])
      {
        DebugBreak(); // halt for inspection if programmer traps a clip
      }
    }
  }
#endif
  //BOOL canEnd = rLink->nodeType == CPN_FILM;
  while (rLink && rLink->nodeType != CPN_FILM)
  {
    CPControl& control = rLink->data.control;
    CPControlData& cdata = rLink->data.control.data;

    switch (control.controlType)
    {
    case CPC_COUNT:
      tLink = rLink->pLinks[control.visitBits?control.visitBits-1:0];
      if (final)
      {
        control.visitBits = ((control.visitBits + 1) % rLink->links)+1;
        // once visited, nonzero.
        // (some only have one link!)
      }
      rLink = tLink;
      break;
    case CPC_BURN:
      // assume it's the one we can never use up.
      k = cdata.burn.links - 1;
      d = psych.ManhattanMetric(cdata.burn.pLinks[k]);

      // search for an unused link with a better match.
      for (i=0; i<cdata.burn.links-1; i++)
      {
        unsigned newD = psych.ManhattanMetric(cdata.burn.pLinks[i]);
        if (newD < d && (control.visitBits & (1<<i)) == 0)
        {
          d = newD;
          k = i;
        }
      }
      if (final && k != unsigned(cdata.burn.links-1)) // it's not the last one
      {
        control.visitBits |= (1<<k);
      }
      assert(k < (unsigned)rLink->links);
      rLink = rLink->pLinks[k];
      break;
    case CPC_SWITCH:
      k = 0;
      d = psych.ManhattanMetric(cdata.burn.pLinks[k]);

      // search for an unused link with a better match.
      for (i=1; i<cdata.burn.links; i++)
      {
        unsigned newD = psych.ManhattanMetric(cdata.burn.pLinks[i]);
        if (newD < d)
        {
          d = newD;
          k = i;
        }
      }
      assert(k < (unsigned)rLink->links);
      rLink = rLink->pLinks[k];
      break;
    case CPC_RELAY:
      tLink = cdata.relayCoil;
      assert(rLink->links == 2);
      assert(tLink->data.control.controlType == CPC_COUNT);
      if (tLink->data.control.visitBits == unsigned(tLink->links))
      {
        // click (well, some of us CAN hear relays).
        rLink = rLink->pLinks[1];
      }
      else
      {
        rLink = rLink->pLinks[0];
      }
      break;
    case CPC_CENSOR:
      // If there are not enough nodes and this branch
      // is not final, returning 0 will disable a path.
      // For example, the prototype disables a path in
      // which protagonist kills the nurse if violence
      // is supposed to be minimized (he still partly
      // strangles her, but then he shows mercy.)
      tLink = 0;
      if (globalOptions.MinViolenceOn())
      {
        if (globalOptions.MinSexOn())
        {
          if (rLink->links>=4)
          {
            tLink = rLink->pLinks[3];
          }
        }
        else
        {
          if (rLink->links>=3)
          {
            tLink = rLink->pLinks[2];
          }
        }
      }
      else if (globalOptions.MinSexOn())
      {
        if (rLink->links >=2)
        {
          tLink = rLink->pLinks[1];
        }
      }
      else // the 'everything goes' case.
      {
        tLink = rLink->pLinks[0];
      }
      rLink = tLink;
      break;
    case CPC_TERMINAL:
      assert(rLink->links==0);
      rLink = 0;
      break;
    case CPC_CHANGEDISK:
      assert(rLink->links != 0);
      return rLink; // this is the specialest special case yet.
    default:
      assert(!"Invalid control type");
      rLink = 0;
      break;
    }
  }
  return rLink;
}

//----------------------------------------------------------------------------
// UnFlatten:
// This function converts all the file offsets in the CPNode to addresses.
// Then it follows the links in the node recursively.
// If a link was already unflattened, it doesn't recursively follow it.
//
void CPNetwork::UnFlatten(CPNode* node)
{
  if (node->links)
  {
    if (!UnFlattenPtr(node->pLinks))
    {
      for (int i=0; i<node->links; i++)
      {
        if (!UnFlattenPtr(node->pLinks[i]))
        {
          UnFlatten(node->pLinks[i]);
        }
      }
    }
  }
  if (node->backLinks)
  {
    if (!UnFlattenPtr(node->pBackLinks))
    {
      for (int i=0; i<node->backLinks; i++)
      {
        if (!UnFlattenPtr(node->pBackLinks[i]))
        {
          UnFlatten(node->pBackLinks[i]);
        }
      }
    }
  }
  // All the links are flattened, and what they point to is flattened.
  switch (node->nodeType)
  {
  case CPN_FILM:
    if (node->data.film.rule == CP_HYPER)
    {
      UnFlattenPtr(node->data.film.data.prompt);
    }
    if (node->data.film.tracks)
    {
      UnFlattenPtr(node->data.film.pTracks);
      int i;
      for (i = 0; i < node->data.film.tracks; i++)
      {
        unsigned j;
        UnFlattenPtr(node->data.film.pTracks[i].pSegments);
        for (j=0; j<node->data.film.pTracks[i].segments; j++)
        {
          UnFlattenPtr(node->data.film.pTracks[i].pSegments[j].prompt);
        }
      }
    }
    break;
  case CPN_CONTROL:
    switch (node->data.control.controlType)
    {
    case CPC_RELAY:
      UnFlattenPtr(node->data.control.data.relayCoil);
      break;
    case CPC_BURN:
    case CPC_SWITCH:
      UnFlattenPtr(node->data.control.data.burn.pLinks);
      break;
    }
    break;
  }
}
//
int CPNetwork::Arc2Vector(void* bf, int handle, StringTab& tab)
{
  FRSRC* bigfile = (FRSRC*)bf;
  int nClipNames;
  bigfile->Read(handle, &nClipNames, sizeof(nClipNames));
  typedef char* CHARPTR;
  char** table = new CHARPTR[nClipNames];
  int i;
  for (i=0; i<nClipNames; i++)
  {
    char buf[80];
    char* cp = buf;
    char ch;
    bigfile->Read(handle, &ch, 1);
    while (ch!=0)
    {
      *cp++ = ch;
      bigfile->Read(handle, &ch, 1);
    }
    *cp = 0;
    table[i] = new char[strlen(buf)+1];
    strcpy((char*)table[i], buf);
  }
  tab.Clear();
  tab.strings = table;
  tab.n = nClipNames;
  return 1;
}
//----------------------------------------------------------------------------
// File2Vector:
// At the current offset into the file:
// int nStrings;
// cstring 0;
// cstring 1;
// ...
// cstring n-1;
// Each cstring is null-terminated, and there is no other indication of its
// length.
//
int                                   // nz if successful
  CPNetwork::File2Vector
    ( void* vtestf                    // FILE* of file to read
    , StringTab& tab                  // table to fill in.
    )
{
  FILE* testf = (FILE*)vtestf;
  int nClipNames;
  fread(&nClipNames, sizeof(nClipNames), 1, testf);
  typedef char* CHARPTR;
  char** table = new CHARPTR[nClipNames];
  int i;
  for (i=0; i<nClipNames; i++)
  {
    char buf[80];
    char* cp = buf;
    int c = getc(testf);
    while (c != EOF && c!=0)
    {
      *cp++ = c;
      c = getc(testf);
    }
    if (c==EOF)
    {
      break;// wha?
    }
    *cp = 0;
    table[i] = new char[strlen(buf)+1];
    strcpy((char*)table[i], buf);
  }
  tab.strings = table;
  tab.n = nClipNames;
  return 1;
}

//-----------------------------------------------------------------------------
// GetNextLeadout:
// Look ahead for the next leadout. 
// If that leadout is adjacent to a visited leadin, 
// look ahead for the leadout which follows.
// Effectively, find the farthest out leadout at the end of a huge 
// sequence of skipped videos.
CPNode*                               // return 0 or a leadout
  CPNetwork::GetNextLeadout
    ( CPNode* node                    // allegedly a leadin
    , const Psych& psych              // the compatibility index
    ) const
{
  CPNode* rp = 0;
  while (node != 0 && node->data.film.type != CP_LEADOUT)
  {
    if (node->data.film.type== CP_LOOP
      ||node->data.film.rule== CP_HYPER
      )
    {
      break; // fah-get it. found a branch instead.
    }
    node = GetCompatibleLink(node->pLinks[0], psych);
    if (node->data.film.type == CP_LEADOUT)
    {
      rp = node;        // consolidate this result
      
      CPNode* trialLead = GetCompatibleLink(node->pLinks[0], psych);
      if (trialLead && trialLead->data.film.type == CP_LEADIN && trialLead->data.film.viewed)
      {
        node = trialLead; // try to extend through another visited leadin
      }
    }
  }
  return rp;
}
