#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "fontbits.h"
#include "cstring.h"
//----------------------------------------------------------
// Private constants, masks, and macros
// First for reference, atari color masks, byte-swapped.
#define ATARI_RED 0x00f8
#define ATARI_BLUE 0xc007
#define ATARI_GREEN 0x3f00
// Macros that convert the atari colors to windows 565 colors
#define A2WRED(x) ((x&0x00f8)<<8)
#define A2WBLU(x) (((x&0x0007)<<2)+((x&0xc000)>>14))
#define A2WGRN(x) ((x&0x3f00)>>3)
// Macro to convert the whole word from atari to windows color.
#define A2W(x) (A2WRED(x)+A2WBLU(x)+A2WGRN(x))

//----------------------------------------------------------
// This function sets up the font from the requested resource id.
// This includes making a dib section to facilitate recoloring the
// bitmap data according to the desired colors.
void 
  FontBits::Rep::SetupResource
    (LPCTSTR name)
{
  HRSRC r = FindResource(0, name, "FONTBITS");
  HGLOBAL gh = LoadResource(0, r);
  fontHdr* pFont = (fontHdr*)LockResource(gh);
  font = *pFont;
  BYTE* bytes = (BYTE*)pFont->data;
  int lineWidth = font.cellWidth*font.nGlyphs;
  int byteCount = lineWidth*font.cellHeight*2;

  rawDib = DibSection(bytes, CSize(lineWidth, font.cellHeight), DibSection::T565);
  readyDib = DibSection(0, CSize(lineWidth*2, font.cellHeight*2), DibSection::T565);

  //CopyMemory(rawDib->GetData(), bytes, byteCount);
  CopyMemory(widths, bytes+byteCount, 64);
  FreeResource(gh); // ?? not listed in win32 hlp.
  // Now "bytes" is a dead variable, and a dangling pointer.
  
  WORD* dibWords = (WORD*)rawDib->GetData();

  // Most bitmaps are top-down, we need bottom up.
  // Atari 16-bit colors are different, we need to fix it.
  WORD* pTop = &dibWords[0];
  WORD* pBot = &dibWords[(font.cellHeight-1)*lineWidth];
  for (int y=0; y<(font.cellHeight/2); y++)
  {
    WORD* tTop = pTop;
    WORD* tBot = pBot;
    for (int x=0; x<1024; x++)
    {
      WORD t = A2W(*tTop);
      *tTop = A2W(*tBot);
      *tBot = t;
      tTop++;
      tBot++;
    }
    pTop = tTop; //pTop+=1024;
    pBot-=1024;
  }
  if (font.cellHeight & 1) // do scanline in middle
  {
    WORD* tTop = pTop;
    for (int x=0; x<1024; x++)
    {
      WORD t = A2W(*tTop);
      *tTop = t;
      tTop++;
    }
  }
  ZeroMemory(colorTrack, sizeof(colorTrack));
  firstGlyph = font.firstAscii;
}

//---------------------------------------------------------------
// DrawString(HDC, string)
// Draw a string, and advances by the built-in width table.
void FontBits::Rep::DrawString
  ( const CDC& ddc
  , const string& text
  )
{
  const char* t = text.c_str();
  while (*t)
  {
    unsigned char tc = toglyph(*t);
    DrawCharacter(ddc, tc);
    cp.x += zoom*widths[tc];
    t++;
  }
}

//---------------------------------------------------------------
// DrawString(HDC, string, const int*)
// Draws a string and advances by the specified character widths.
// Stops at the end of the string.
void FontBits::Rep::DrawString
  ( const CDC& ddc                           // device context
  , const string& text                // string to draw
  , const int* deltas                 // widths to use.
  )
{
  DrawString(ddc, text.c_str(), deltas, text.length());
}

//---------------------------------------------------------------
// DrawString(HDC, const char*, const int*, int)
// Draws a string, advances by the specified widths. 
// Draws the specified number of characters.
void FontBits::Rep::DrawString
  ( const CDC& ddc                           // device context
  , const char* text                  // characters to draw.
  , const int* deltas                 // widths to use.
  , int nChars                        // number of each.
  )
{
  for (int i=0; i<nChars; i++)
  {
    DrawCharacter(ddc, toglyph(text[i]));
    cp.x += zoom*deltas[i];
  }
}

//---------------------------------------------------------------
// MeasureString:
// Returns the total size of the string.
CSize 
  FontBits::Rep::MeasureString
    (const string& text) const
{
  CSize sz;
  sz.cy = zoom*19;
  const char* t = text.c_str();
  while (*t)
  {
    sz.cx += zoom*widths[toglyph(*t)];
    t++;
  }
  return sz;
}

//---------------------------------------------------------------
// GetWidths:
// Gets the widths of each character.
// Returns TRUE if all went well.
// Returns FALSE if a character was out of bounds.
BOOL 
  FontBits::Rep::GetWidths
    ( const string& text
    , int* widths
    ) const
{
  return GetWidths(text.c_str(), widths, text.length());
}
//---------------------------------------------------------------
// GetWidths:
// Gets the widths of each character.
// Returns TRUE if all went well.
// Returns FALSE if a character was out of bounds.
BOOL 
  FontBits::Rep::GetWidths
    ( const char* text
    , int* widths
    , int nChars
    ) const
{
  int rv=TRUE;
  const char* t= text;
  while (*t)
  {
    int w;
    if (*t-firstGlyph < font.nGlyphs)
    {
      w = zoom*widths[toglyph(*t)];
    }
    else
    {
      rv = FALSE;
      w = 0;
    }
    t++;
  }
  return rv;
}

//----------------------------------------------------------------------------
// GetAllWidths:
// Return the specified number of widths, starting with the
// width of the first glyph.
// If the number of widths requested is the number of widths
// available, the function returns true.
// The function copies no more than asked, and no more than can be copied.
BOOL                                  // T if you got what you asked
  FontBits::Rep::GetAllWidths
    ( unsigned* widthsOut             // destination
    , unsigned nWidths                // how many
    ) const
{
  unsigned toCopy = min(nWidths, font.nGlyphs);
  for (unsigned i=0; i<toCopy; i++)
  {
    *widthsOut++ = zoom*widths[i];
  }
  return toCopy != nWidths;
}

//---------------------------------------------------------------
// ReadyCharacter:
// This function prepares the cached glyph image.
// The character is recolored if the color is not the one recorded in
// the tracking array, or if the zoom is not right.
// Because of the color tracking array, this function is slightly
// non-const. BUT, I want it to be const.
void FontBits::Rep::ReadyCharacter
  (int glyph) const                   // The glyph to prepare.
{
  if (color != colorTrack[glyph].color || zoom != colorTrack[glyph].zoom)
  {
    if (zoom==1)
    {
      WORD* tlws = ((WORD*)(rawDib->GetData())) + glyph*font.cellWidth;
      WORD* tlwd = ((WORD*)(readyDib->GetData())) + glyph*font.cellWidth*2;
      for (int y=0; y<font.cellHeight; y++)
      {
        for (int x=0; x<font.cellWidth; x++)
        {
          tlwd[x] = tlws[x]&color;
        }
        tlws+=1024;
        tlwd+=2048;
      }
    }
    else if (zoom==2)
    {
      WORD* tlws = ((WORD*)(rawDib->GetData())) + glyph*font.cellWidth;
      WORD* tlwd = ((WORD*)(readyDib->GetData())) + glyph*font.cellWidth*2;
      for (int y=0; y<font.cellHeight; y++)
      {
        // colorize and stretch one scanline of character.
        for (int x=0; x<font.cellWidth; x++)
        {
          tlwd[2*x] = tlws[x]&color;
          tlwd[2*x+1] = tlwd[2*x];
        }
        // double the scanline.
        CopyMemory(tlwd+2048, tlwd, font.cellWidth*2*sizeof(WORD));

        tlws+=1024;
        tlwd+=4096; // advance dest by 2 scanlines.
      }
    }
    cacheTrack* sp = const_cast<cacheTrack*>(colorTrack+glyph);
    sp->color = color;
    sp->zoom = zoom;
  }
}
//---------------------------------------------------------------
// DrawCharacter:
// This function draws a character and advances the current point
// by the width of the character.
// The character is recolored if the color is not the one recorded in
// the tracking array.
// Because of the color tracking array, this function is slightly
// non-const. 
void FontBits::Rep::DrawCharacter
  ( const CDC& ddc             // the dc to draw to.
  , int glyph                  // the glyph index
  ) const
{
  ReadyCharacter(glyph);
  switch (zoom)
  {
  case 1:
    readyDib->Blit( ddc
                  , cp
                  , CRect( 2*glyph*font.cellWidth+font.cellWidth-widths[glyph]
                         , font.cellHeight
                         , 2*glyph*font.cellWidth+font.cellWidth
                         , 2*font.cellHeight
                         )
                   );
    break;
  case 2:
    readyDib->Blit( ddc
                  , cp
                  , CRect( 2*(glyph*font.cellWidth+font.cellWidth-widths[glyph])
                         , 0
                         , 2*(glyph*font.cellWidth+font.cellWidth)
                         , 2*font.cellHeight
                         )
                  );
    break;
  default:
    DebugBreak();
    break;
  }
}

